
// 
// Copyright (c) 2011 Jeff Flinn
// 
// Distributed under the Boost Software License, Version 1.0. (See accompanying 
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt) 
// 

#ifndef BOOST_PROCESS_POSIX_EXECUTOR_HPP
#define BOOST_PROCESS_POSIX_EXECUTOR_HPP

#include <boost/process/config.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/fusion/algorithm/iteration/for_each.hpp>
#include <boost/process/posix/initializers/initializer.hpp>

#include <fstream>
#include <vector>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

//////////////////////////////////////////////////////////////////
#define BOOST_PROCESS_EXECUTOR_METHOD(fnc)                        \
                                                                  \
template<typename INITIALIZERS> void fnc(const INITIALIZERS& izs) \
{                                                                 \
    boost::fusion::for_each(izs, initializer::lazy::fnc(*this));  \
}
//////////////////////////////////////////////////////////////////

namespace boost { namespace process { namespace posix {

    struct executor 
    {
        typedef boost::filesystem::path      path;
        typedef char                    char_type;
        typedef pid_t                    pid_type;

        executor() : m_exe(), m_arg_ptrs(), m_env_vars_ptrs(0), m_fd_map(), m_pid(-1), m_errno(0) {}
        
        template<typename INITIALIZERS> void exec(const INITIALIZERS& izs)
        {
            pre_fork_parent(izs);
            
            switch(m_pid = fork())
            {
                case  0:   post_fork_child (izs); return; 
                default:   post_fork_parent(izs); return; 
                case -1: failed_fork_parent(izs); return;
            }
        }

        template<typename INITIALIZERS> void post_fork_child(const INITIALIZERS& izs)
        {
            boost::fusion::for_each(izs, initializer::lazy::post_fork_child(*this));
            
            m_fd_map.set_child_std_fds();
            
            parent_file_descriptor_map::close_unused_fds();
        
            //execv("/xApplications/TextEdit.app/Contents/MacOS/TextEdit", 0);//m_arg_ptrs.ptr());//m_env_vars_ptrs); // doesn't return on success
            execve(m_exe.c_str(), m_arg_ptrs.ptr(), m_env_vars_ptrs); // doesn't return on success
            
            failed_exec_child(izs);
        }
        
        BOOST_PROCESS_EXECUTOR_METHOD(   pre_fork_parent)
        BOOST_PROCESS_EXECUTOR_METHOD(  post_fork_parent)
        BOOST_PROCESS_EXECUTOR_METHOD(failed_fork_parent)
        BOOST_PROCESS_EXECUTOR_METHOD(failed_exec_child )
    
        struct arg_ptrs
        {
            arg_ptrs() : m_arg_ptrs(2, (char*)0) {} // front is for app name, back is trailing null 
        
            void assign_app(char_type* n) { m_arg_ptrs.front() = n; }
            void  push_back(char_type* a) { m_arg_ptrs. back() = a; m_arg_ptrs.push_back(0); }
            
            char_type* const* ptr() const { return &m_arg_ptrs[0]; }
        
        private:
            typedef std::vector<char_type*> arg_ptrs_vec;
            
            arg_ptrs_vec m_arg_ptrs;
        };
        
        void set_std_in (int fd) { m_fd_map.set_parent_fds(fd); }
        void set_std_out(int fd) { m_fd_map.set_parent_fds(fd); }
        void set_std_err(int fd) { m_fd_map.set_parent_fds(fd); }
       
        struct parent_file_descriptor_map
        {
            parent_file_descriptor_map()
            : dupd_parent_std_in (-1)
            , dupd_parent_std_out(-1)
            , dupd_parent_std_err(-1)
            {}

            void set_parent_fds(int fd) // call from child process
            {
                switch(fd)
                {
                    case  0: dupd_parent_std_in  = ensure_user_fd_index(fd); return; 
                    case  1: dupd_parent_std_out = ensure_user_fd_index(fd); return; 
                    case  2: dupd_parent_std_err = ensure_user_fd_index(fd); return; 
                    default:                                                 return; 
                }
            }

            void set_child_std_fds()
            {
                if(-1 != dupd_parent_std_in ) { dup2(dupd_parent_std_in ,  STDIN_FILENO); } else { close( STDIN_FILENO); }
                if(-1 != dupd_parent_std_out) { dup2(dupd_parent_std_out, STDOUT_FILENO); } else { close(STDOUT_FILENO); }
                if(-1 != dupd_parent_std_err) { dup2(dupd_parent_std_err, STDERR_FILENO); } else { close(STDERR_FILENO); }
            }
            
            static void close_unused_fds()
            {
                for(std::size_t i = 3 ; i < 1000 ; ++i ) close(i);
            }

            static int ensure_user_fd_index(int fd_index)
            {
                //BOOST_STATIC_ASSERT( STDIN_FILENO == 0)
                //BOOST_STATIC_ASSERT(STDOUT_FILENO == 1)
                //BOOST_STATIC_ASSERT(STDERR_FILENO == 2)

                const int user_fileno_begin = 3;

                int user_fd_index = fcntl(fd_index, F_DUPFD, user_fileno_begin);

                return user_fd_index;
            }

            int dupd_parent_std_in ;        
            int dupd_parent_std_out;        
            int dupd_parent_std_err;        
        };

        path                       m_exe;
        arg_ptrs                   m_arg_ptrs;
        char_type**                m_env_vars_ptrs;
        parent_file_descriptor_map m_fd_map;
        pid_type                   m_pid;
        int                        m_errno;
        
    private:
                   executor(const executor&);
        executor& operator=(const executor&);
    };

}}}

#undef BOOST_PROCESS_EXECUTOR_METHOD

#endif
